{ File   : SND2WAV.dpr
  Author : Deniz Oezmen
  Created: 2007-01-14
  Changed: 2007-01-20

  The Last Express SND converter.
}
program SND2WAV;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

const
  BlockSize = 739;

type
  TSNDHeader   = packed record
                   Size     : Longword;
                   NumBlocks: Word;
                 end;

  TBlockHeader = packed record
                   Sample,
                   Index : Smallint;
                 end;

  TBuffer      = array of Byte;

var
  SNDFile  : file;
  WAVHeader: packed array[0..43] of Char = (
      'R',  'I',  'F',  'F', #$00, #$00, #$00, #$00,  'W',  'A',  'V',  'E',
      'f',  'm',  't',  ' ', #$10, #$00, #$00, #$00, #$01, #$00, #$01, #$00,
     #$44, #$ac, #$00, #$00, #$88, #$58, #$01, #$00, #$02, #$00, #$10, #$00,
     'd',  'a',  't',  'a',  #$00, #$00, #$00, #$00);

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenSNDFile(var SNDFile: file; SNDName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(SNDFile, SNDName);

  {$I-}
  Reset(SNDFile, 1);
  {$I+}
  OpenSNDFile := (IOResult = 0)
end;

function ReadHeader(var SNDFile: file; out SNDHeader: TSNDHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(SNDFile, SNDHeader, SizeOf(SNDHeader), BytesRead);

  Result := (BytesRead = SizeOf(SNDHeader)) and
    (SNDHeader.Size = SNDHeader.NumBlocks * BlockSize)
end;

procedure IMA_GetNewSample(var Sample, Idx: Smallint; SampleCode: Byte);
const
  StepTab: array[0..88] of Smallint =
    ($0007, $0008, $0009, $000a, $000b, $000c, $000d, $000e,
     $0010, $0011, $0013, $0015, $0017, $0019, $001c, $001f,
     $0022, $0025, $0029, $002d, $0032, $0037, $003c, $0042,
     $0049, $0050, $0058, $0061, $006b, $0076, $0082, $008f,
     $009d, $00ad, $00be, $00d1, $00e6, $00fd, $0117, $0133,
     $0151, $0173, $0198, $01c1, $01ee, $0220, $0256, $0292,
     $02d4, $031c, $036c, $03c3, $0424, $048e, $0502, $0583,
     $0610, $06ab, $0756, $0812, $08e0, $09c3, $0abd, $0bd0,
     $0cff, $0e4c, $0fba, $114c, $1307, $14ee, $1706, $1954,
     $1bdc, $1ea5, $21b6, $2515, $28ca, $2cdf, $315b, $364b,
     $3bb9, $41b2, $4844, $4f7e, $5771, $602f, $69ce, $7462,
     $7fff);
  IndexTab: array[0..15] of Smallint =
    ( -1, -1, -1, -1,  2,  4,  6,  8,
      -1, -1, -1, -1,  2,  4,  6,  8);
var
  Samp0,
  Idx0,
  SampX,
  Diff : Longint;
begin
  // expand parameters' value range for computations
  Samp0 := Longint(Sample);
  Idx0  := Longint(Idx);
  SampX := Longint(SampleCode);

  // calculate difference
  Diff := 0;
  if SampX and $04 <> 0 then
    Diff := Diff + StepTab[Idx0];
  if SampX and $02 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 1);
  if SampX and $01 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 2);
  Diff := Diff + (StepTab[Idx0] shr 3);
  if SampX and $08 <> 0 then
    Diff := -Diff;

{  alternate method:
  Diff := (StepTab[Idx0] * (SampX and $00000007) shl 1 + 1) shr 3;
  if SampX and $08 <> 0 then
    Diff := -Diff;}

  // calculate new sample
  Samp0 := Samp0 + Diff;
  if Samp0 > 32767 then
    Samp0 := 32767
  else if Samp0 < -32768 then
    Samp0 := -32768;

  // calculate new step table index
  Idx0 := Idx0 + IndexTab[SampX];
  if Idx0 > 88 then
    Idx0 := 88
  else if Idx0 < 0 then
    Idx0 := 0;

  // write values back to parameters
  Sample := Smallint(Samp0);
  Idx    := Byte(Idx0)
end;

function IMA_ConvertBlock(BlockSize: Longint; const InBuffer: array of Byte;
  var OutBuffer: array of Smallint): Boolean;
var
  i,
  NumSamples,
  InPos,
  OutPos     : Longint;
  BlockHeader: TBlockHeader;
begin
  Result := False;

  NumSamples := BlockSize - SizeOf(TBlockHeader);

  if Length(OutBuffer) < NumSamples * 2 then
  begin
    WriteLn(Format(
      '  Internal error: Insufficient buffer size (%d needed, %d allocated).',
      [NumSamples * 2, Length(OutBuffer)]));
    Exit
  end;

  // get block header
  Move(InBuffer[0], BlockHeader, SizeOf(BlockHeader));

  if BlockHeader.Index > 88 then
  begin
    WriteLn(Format('  Warning: Index out of range (got %d, expected <= 88).',
      [BlockHeader.Index]));
    Exit
  end;

  InPos := SizeOf(BlockHeader);
  OutPos := 0;

  // decode block
  with BlockHeader do
    for i := 1 to NumSamples do
    begin
      // high nibble
      IMA_GetNewSample(Sample, Index, InBuffer[InPos] shr 4);
      OutBuffer[OutPos] := Sample;

      // low nibble
      IMA_GetNewSample(Sample, Index, InBuffer[InPos] and $0f);
      OutBuffer[OutPos + 1] := Sample;

      Inc(InPos);
      Inc(OutPos, 2)
    end;

  Result := True
end;

procedure Convert_IMA_ADPCM(SNDName: string; const SNDHeader: TSNDHeader);
var
  i           : Longint;
  SampPerBlock,
  NumSamples,
  UncompSize  : Longword;
  InBuffer    : TBuffer;
  OutBuffer   : array of Smallint;
  WAVFile     : file;
begin
  SNDName := Prefix(SNDName) + '.wav';

  // set 4-bit IMA_ADPCM parameters (mono, 44100 Hz)
  SampPerBlock := BlockSize - SizeOf(TBlockHeader);
  NumSamples := SNDHeader.NumBlocks * SampPerBlock;

  // prepare wave header
  UncompSize := NumSamples * 4;
  Move(UncompSize, WAVHeader[40], 4);

  Inc(UncompSize, 32);
  Move(UncompSize, WAVHeader[4], 4);

  // open output file and write wave header
  FileMode := fmOpenReadWrite;
  Assign(WAVFile, SNDName);
  {$I-}
  Rewrite(WAVFile, 1);
  BlockWrite(WAVFile, WAVHeader[0], SizeOf(WAVHeader));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  Write error.');
    Exit
  end;

  SetLength(InBuffer, BlockSize);
  SetLength(OutBuffer, SampPerBlock * 2);

  for i := 1 to SNDHeader.NumBlocks do
  begin
    // read next block
    {$I-}
    BlockRead(SNDFile, InBuffer[0], BlockSize);
    {$I+}
    if IOResult <> 0 then
    begin
      if Eof(SNDFile) then
        WriteLn('  Warning: Premature end of file; output might be damaged.')
      else
        WriteLn('  Read error.');
      Break
    end;

    // convert block
    if not IMA_ConvertBlock(BlockSize, InBuffer, OutBuffer) then
    begin
      WriteLn(
        '  Warning: Decompression error (skipping); output might be damaged.');
      Continue
    end;

    // write converted block
    {$I-}
    BlockWrite(WAVFile, OutBuffer[0], Length(OutBuffer) * 2);
    {$I+}
    if IOResult <> 0 then
    begin
      WriteLn('  Write error.');
      Break
    end
  end;

  Finalize(InBuffer);
  Finalize(OutBuffer);

  Close(WAVFile)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  SNDName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
  SNDHeader: TSNDHeader;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      SNDName := Dir + FileName;

      WriteLn(SNDName);

      // if this file is accessible, analyze it
      if OpenSNDFile(SNDFile, SNDName) then
      begin
        if ReadHeader(SNDFile, SNDHeader) then
          Convert_IMA_ADPCM(SNDName, SNDHeader)
        else
          WriteLn('  Not a valid SND file.');

        Close(SNDFile)
      end
      else
        WriteLn('  Not accessible.')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(FilePara, ' not found')
end;

begin
  if ParamCount >= 1 then
    ParseFiles(ParamStr(1))
  else
    WriteLn('Incorrect number of parameters.')
end.
